/*
 * archive.c - part of procsms application
 *
 * Copyright (C) 2003 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <turbo/turbo.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "archive.h"
#include "procsms.h"

#ifdef CONFIG_ARCHIVE

/* *INDENT-OFF* */

lc_char PROGMEM lc_Archive[]={
        LC_EN("SMS Archive")
        LC_CZ("SMS Archiv")
        LC_END
};

static lc_char PROGMEM lc_On[]={
        LC_EN("Activate")
        LC_CZ("Zapnout")
        LC_END
};

static lc_char PROGMEM lc_Off[]={
        LC_EN("Deactivate")
        LC_CZ("Vypnout")
        LC_END
};

static lc_char PROGMEM lc_New_Folder[]={
        LC_EN("New Folder")
        LC_CZ("Nova slozka")
        LC_END
};

static lc_char PROGMEM lc_Not_Empty[]={
        LC_EN("Folder Not Empty")
        LC_CZ("Slozka neni prazdna")
        LC_END
};

static lc_char PROGMEM lc_New_Rule[]={
        LC_EN("New Rule")
        LC_CZ("Nove pravidlo")
        LC_END
};

static lc_char PROGMEM lc_Move[]={
        LC_EN("Move")
        LC_CZ("Presunout")
        LC_END
};

static lc_char PROGMEM lc_Drop[]={
        LC_EN("Drop")
        LC_CZ("Vlozit")
        LC_END
};

static lc_char PROGMEM lc_Folder_name[]={
        LC_EN("Folder Name:")
        LC_CZ("Nazev slozky:")
        LC_END
};

static lc_char PROGMEM lc_Delete_All[]={
        LC_EN("Delete Messages")
        LC_CZ("Smazat zpravy")
        LC_END
};

static lc_char PROGMEM lc_Delete_Folder[]={
        LC_EN("Delete Folder")
        LC_CZ("Smazat slozku")
        LC_END
};

static lc_char PROGMEM lc_Delete_Msg[]={
        LC_EN("Delete Message")
        LC_CZ("Smazat zpravu")
        LC_END
};

static lc_char PROGMEM lc_Delete_Rule[]={
        LC_EN("Delete Rule")
        LC_CZ("Smazat prav.")
        LC_END
};

static lc_char PROGMEM lc_Rules[]={
        LC_EN("Rules")
        LC_CZ("Pravidla")
        LC_END
};

/* *INDENT-ON* */

#define RULE_MSISDN	13
#define RULE_POS	3
#define RULE_LEN	(RULE_MSISDN+RULE_POS)

static u8 PROGMEM t_up[] = "..";
static u8 PROGMEM t_sms[] = "sms";
static u8 PROGMEM t_incoming[] = "Inbox";
static u8 PROGMEM t_r[] = ".rules";
static u8 PROGMEM t_links[] = ".dirs";
u8 PROGMEM t_ext_sms[] = "sms";

u8 sprint_date (u8 * t, u8 * h, u8 max_len)
{
  u8 i = 0;

  if (i >= max_len)
    return i;
  h[i] = (t[2] & 0x0F) + '0';
  i++;
  if (i >= max_len)
    return i;
  h[i] = (t[2] >> 4) + '0';
  i++;
  if (i >= max_len)
    return i;
  h[i] = '.';
  i++;
  if (i >= max_len)
    return i;
  h[i] = (t[1] & 0x0F) + '0';
  i++;
  if (i >= max_len)
    return i;
  h[i] = (t[1] >> 4) + '0';
  i++;
  if (i >= max_len)
    return i;
  h[i] = '.';
  i++;
  if (i >= max_len)
    return i;
  h[i] = (t[0] & 0x0F) + '0';
  i++;
  if (i >= max_len)
    return i;
  h[i] = (t[0] >> 4) + '0';
  i++;
  if (i >= max_len)
    return i;
  h[i] = ' ';
  i++;
  if (i >= max_len)
    return i;
  h[i] = (t[3] & 0x0F) + '0';
  i++;
  if (i >= max_len)
    return i;
  h[i] = (t[3] >> 4) + '0';
  i++;
  if (i >= max_len)
    return i;
  h[i] = ':';
  i++;
  if (i >= max_len)
    return i;
  h[i] = (t[4] & 0x0F) + '0';
  i++;
  if (i >= max_len)
    return i;
  h[i] = (t[4] >> 4) + '0';
  i++;
  return i;
}

void seek_sms ()
{
  b32 sms;
  b32 inc;
  b32 rules;
  Inode *ino = inode ();

  dbsp (">SEEK_SMS\n");
  db_stack ();
  cd_i (root ());
  sms = cd (t_sms);
  if (sms < 0)
  {
    dbsp ("S1\n");
// /s
    sms = create_dir (root (), t_sms, SORT_APPEND);
    dbsp ("S2 ");
    dbih (fs_ino_block (sms));
    dbch (fs_ino_index (sms));
    dbc ('\n');
    df_print (fs_ino_block (sms));

// /s/incoming
    inc = create_dir (sms, t_incoming, SORT_SMS);
    dbsp ("S6 ");
    dbih (fs_ino_block (inc));
    dbch (fs_ino_index (inc));
    dbc ('\n');
    df_print (fs_ino_block (inc));

// /s/.rules
    rules = create_file (sms, t_r, SORT_SMS);
    dbsp ("S5 ");
    dbih (fs_ino_block (rules));
    dbch (fs_ino_index (rules));
    dbc ('\n');
    df_print (fs_ino_block (rules));

    rules = create_file (sms, t_links, SORT_SMS);

// /s/incoming/.rules
    rules = create_dir (inc, t_r, SORT_SMS);
    dbsp ("S7 ");
    dbih (fs_ino_block (rules));
    dbch (fs_ino_index (rules));
    dbc ('\n');
    df_print (fs_ino_block (rules));

    cd_i (sms);
  }

  dbsp ("S8\n");
  glob_mem->sms = pwd ();
  get_inode (ino, glob_mem->sms);
  glob_mem->smi = fs_ino_p (ino->content, ino->content_index);	//inbox
  glob_mem->rules = mc_find (0, t_r, FS_FIND_STRING);
  glob_mem->links = mc_find (0, t_links, FS_FIND_STRING);
  dbsp ("S88 ");
  dbih (fs_ino_block (glob_mem->rules));
  dbch (fs_ino_index (glob_mem->rules));
  dbc ('\n');
  dbsp ("<SEEK_SMS\n");
}

#define MSG_NAME_BRIEF		0x01
#define MSG_NAME_MSISDN		0x02
#define MSG_NAME_PROTECTED	0x04
#define MSG_NAME_CRYPTED	0x08
#define MSG_NAME_U		0x10

void msg_name (u8 name[INODE_TEXT_LEN], u8 type, u8 * msg)
{
  u8 *t;
  u8 l;
  u8 k, n;
  u8 *a1;
  u8 dcs;
  u8 len = 2;

  dbsp (">msg_name ");
  dbch (type);
  dbc ('\n');

  bzero (name, INODE_TEXT_LEN);

  if (msg != NULL)
  {
    if ((*msg & 0x03) != 0x00)	//DELIVER
      name[0] = '<';
    else
    {
      if ((type & MSG_NAME_U) == MSG_NAME_U)
	name[0] = '*';
      else
	name[0] = '>';
    }
  }

  name[1] = ' ';

#ifdef CONFIG_ARCHIVE_EXT
  if ((type & MSG_NAME_CRYPTED) == MSG_NAME_CRYPTED)
    name[1] = 'C';

  if ((type & MSG_NAME_PROTECTED) == MSG_NAME_PROTECTED)
  {
    memcpy (&name[2], t_protected, strlen (t_protected));
    len += strlen (t_protected);
  }
  else
#endif

  if ((type & MSG_NAME_MSISDN) == MSG_NAME_MSISDN)
  {
    t = tpdu_seek (msg, T_SMS_AA);
    a1 = get_adn (t, MSISDN_SMS);
    l = *t;
    if (a1 != NULL)
    {
      dbsp ("ADN FOUND\n");
      l = 0;
      while (a1[l] != 0xFF && a1[l] != '\0' && l <= adn_len () - 14)
	l++;
      if (l > INODE_TEXT_LEN - 3)
	l = INODE_TEXT_LEN - 3;
      len += l;
      for (k = 0; k < l; k++)
	name[k + 2] = a1[k];
    }
    /*
       else
       {
       dbsp("ADN NOT_FOUND\n");
       if(l>INODE_TEXT_LEN-4)l=INODE_TEXT_LEN-4;
       len+=l;
       a1=msisdn2str(t,MEM_R,MSISDN_SMS,MEM_R);
       for(k=0;a1[k]!='\0'&& k<INODE_TEXT_LEN-4;k++)name[2+k]=a1[k];
       }
     */
    free (a1);
  }
  else if ((type & MSG_NAME_BRIEF) == MSG_NAME_BRIEF)
  {
    t = tpdu_seek (msg, T_SMS_DCS);
    dcs = *t;
    t = tpdu_seek (msg, T_SMS_UDL);
    l = *t;
    if (l > 0)
    {
      if (l > INODE_TEXT_LEN - 3)
	l = INODE_TEXT_LEN - 3;
      len += l;

      for (k = 0; k < l; k++)
      {
	if (dcs_is_7b (dcs))
	{
	  for (n = 0; n < 7; n++)
	  {
	    name[2 + k] |= get_bit (&t[1], k * 7 + n) << n;
	  }
	}
	else
	  name[2 + k] = t[1 + k];
      }
    }
  }

  name[len] = '\0';
}

void archive_store_sms (u8 * data, u8 flag, u8 rec_p1, u8 rec_p2)
{
  b32 b, w, f, rules_len;
  Inode *ino = inode ();
  u8 *t;
  u8 *s;
  u8 *name;
  u8 i;
  u16 k;
  u8 msg_type = MSG_NAME_U;
  u8 nuhdi;
  u8 found = 0;
  MCFILE *fd, *frules;

#define MAX_SMS_LEN	200
  dbsp (">store_sms\n");
  dbsp ("FLAG ");
  dbch (flag);
  dbc ('\n');

  if (flag == SMS_CRYPTED)
    msg_type |= MSG_NAME_CRYPTED;

  if (rec_p1 != 0xFF)
  {
    // from ME to SIM
    dbsp ("STORE_TO_SIM\n");

    //data[0]=0;for(i=1;i<0xB0;i++)data[i]=0xFF; //delete

    i = data[0];
    sim (ME_CMD_UPDATE_RECORD, rec_p1, rec_p2, 0xB0, data);
    if (glob_mem->activ == 0)
      return;
    if (mc_detect () != MC_OK)
      return;

    //SMS RECEIVED, TO BE READ
    //SMS SENT TO NET, SMS STATUS NOT REQ
    //SMS SENT TO NET, SMS STATUS REQ, NOT REC
    if ((i & 0x01) == 0x00)
      return;			//DELETED
    if ((i & 0x07) == 0x01)
      return;			//RECEIVED AND READ
    if ((i & 0x07) == 0x03)
      goto X1;			//RECEIVED, TO BE READ
    if ((i & 0x07) == 0x07)
      return;			//TO BE SENT
    if ((i & 0x18) != 0x00 && (i & 0x18) != 0x08)
      return;			//SENT TO NET, SMS STATUS NOT REQ or SMS STATUS REQ, NOT RECEIVED
    // &0x18==0x10 STATUS REQ, RECEIVED, NOT STORED IN SMSR
    // &0x18==0x18 STATUS REQ, RECEIVED, STORED IN SMSR

  X1:
    dbsp ("READ_FROM_SIM\n");
    sim (ME_CMD_READ_RECORD, rec_p1, rec_p2, 0xB0, data);
    s = data;
    s++;			//status
    i = *s;
    s += i + 1;			//skip SMSC
  }
  else
    s = data;

  if (((*s) & 0x03) != 0x00)
    return;			// not DELIVER

  dbsp ("OK, valid MSG for ARCHIVE\n");

  if (((*s) & 0x40) == 0)
    nuhdi = 1;
  else
    nuhdi = 0;

  name = malloc (INODE_TEXT_LEN);
  if (name == NULL)
    return;

  t = tpdu_seek (s, T_SMS_AA);
  w = pwd ();
  seek_sms ();
  rules_len = get_len (glob_mem->rules);
  rules_len /= RULE_LEN;

  if (rules_len)
  {
    u8 *buf = ino;

    frules = mc_open (glob_mem->rules, FILE_READ);
    for (k = 0; k < rules_len; k++)
    {
      mc_read (frules, k * RULE_LEN, buf, RULE_LEN);
      if (buf[0] != 0 && msisdncmp (buf, MSISDN_ADN, t, MSISDN_SMS))
      {
	dbsp ("MSISDN FOUND\n");

	found = 1;
	if (nuhdi)
	  msg_type |= MSG_NAME_BRIEF;

	b =
	  fs_ino_p ((u16) buf[RULE_MSISDN] |
		    (buf[RULE_MSISDN + 1] << 8), buf[RULE_MSISDN + 2]);
	msg_name (name, msg_type, s);
	f = create_file (b, name, SORT_SMS);
	if (f < 0)
	{
	  mc_close (frules);
	  goto X_END;
	}
	mc_set_type (f, t_ext_sms);

	fd = mc_open (f, FILE_APPEND);
	mc_append (fd, s, MAX_SMS_LEN);
	mc_close (fd);
      }
    }
    mc_close (frules);
  }

  if (found == 0)
  {
    dbsp ("MSISDN NOT FOUND\n");
    msg_type |= MSG_NAME_MSISDN;
    msg_name (name, msg_type, s);
    f = create_file (glob_mem->smi, name, SORT_SMS);	// /s/i/
    if (f < 0)
      goto X_END;
    mc_set_type (f, t_ext_sms);

    fd = mc_open (f, FILE_APPEND);
    mc_append (fd, s, MAX_SMS_LEN);
    mc_close (fd);
  }
X_END:
  free (name);
  cd_i (w);

  dbsp ("<store_sms\n");
}

extern SNodeP archive_n;
extern SNodeP archive_n_new_folder;
extern SNodeP archive_n_delete_folder;
extern SNodeP archive_n_delete_msg;
extern SNodeP archive_n_delete_rule;
extern SNodeP archive_n_delete_all;
extern SNodeP archive_n_msg_view;
extern SNodeP archive_n_msg_reply;
extern SNodeP archive_n_msg_forward;
extern SNodeP archive_n_new_rule;
extern SNodeP archive_n_rules;
extern SNodeP archive_n_move;
extern SNodeP archive_n_drop;
extern SNodeP archive_n_on;
extern SNodeP archive_n_off;

u8 archive_df (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    Inode *ino = inode ();
    u8 i;

    dbsp (">archive_df_create\n");
    glob_mem->tmp_b = ctx->eD;
    get_inode (ino, glob_mem->tmp_b);
    glob_mem->parent_pwd = ino_parent (ino);
    //!!! if not found!!!
#ifdef DEBUG
    dbsp ("W ");
    dbih (ctx->eD);
    dbc ('\n');
    for (i = 0; i < sizeof (Inode); i++)
    {
      dbch (*(u8 *) ((u8 *) ino + i));
      dbc (' ');
    }
    dbc ('\n');
#endif
    if (ino->type == INODE_TYPE_DIR)
    {
      cd_i (glob_mem->tmp_b);
      glob_mem->is_rules = 0;

      glob_mem->_pwd = pwd ();
      ctx->n_default = 2;
      // "new folder"
      // "delete all messages"
      // "rules"
      // "delete folder"
      // "drop"

      if (glob_mem->drop)
	spider_append_r (ctx, &archive_n_drop);

      spider_append_r (ctx, &archive_n_delete_all);

      if (glob_mem->tmp_b != glob_mem->smi)
      {
	spider_append_r (ctx, &archive_n_rules);
	spider_append_r (ctx, &archive_n_delete_folder);
	spider_append_r (ctx, &archive_n_new_folder);

	ctx->n_default = 5;
      }

    }
    else
      //if(ino->type==INODE_TYPE_FILE)
    {
      if (!glob_mem->is_rules)
      {
	spider_append_r (ctx, &archive_n_delete_msg);
	spider_append_r (ctx, &archive_n_move);
	spider_append_r (ctx, &archive_n_msg_forward);
	spider_append_r (ctx, &archive_n_msg_reply);
	spider_append_r (ctx, &archive_n_msg_view);
      }
      else
	spider_append_r (ctx, &archive_n_delete_rule);

      ctx->n_default = 0;
    }
    dbsp ("<archive_df_create\n");
  }
  else if (action == APP_LEAVE)
  {
    dbsp ("archive_df_free\n");
    spider_clear_r (ctx);
  }
  return APP_OK;
}

u8 *file_2_buf_B (u32 file, u8 ret_key, u8 * encrypted)
{
  b16 j;
  u8 *res, *t, *a1;
  Inode *ino = inode ();
  u8 i, k, l;
  u8 *v = NULL;
  u8 ret;
  u8 file_len;
  MCFILE *fd;

  dbsp ("file_2_buf_B\n");

  file_len = j = get_len (file);
  dbsp ("FLEN: ");
  dbih (j);
  dbc ('\n');
  *encrypted = 0;

  v = malloc (j);
  if (v == NULL)
  {
    i = perror (ERR_NO_RAM);
    if (i == APP_END)
      return ENULL;
    return NULL;
  }

  fd = mc_open (file, FILE_READ);
  mc_read (fd, 0, v, j);
  mc_close (fd);

  //masked
  if (check_mask (v, 0xff) == 0)
  {
    if (check_crypted (v))
    {
      // crypted
      *encrypted = 1;

      i = decrypt_msg (v);
      if (i != APP_OK)
      {
	free (v);
	if (i == APP_END)
	  return ENULL;
	return NULL;
      }
    }
  }

  i = *v;
  if (i & 0x40)
  {
    u16 bits, b1;

    // TP_UHDI
    dbsp ("UHDI\n");
    a1 = tpdu_seek (v, T_SMS_DCS);
    j = *a1;
    a1 = tpdu_seek (v, T_SMS_UDL);
    l = *a1;
    a1++;
    i = *a1;			//UDHL
    if (dcs_is_7b (j))
    {
      dbsp ("7B\n");
      j = 7 - ((1 + i) * 8) % 7;	//padding, fill bits
      if (j == 7)
	j = 0;
      dbsp ("JP ");
      dbih (j);
      dbc ('\n');
      j += (1 + i) * 8;		// header in bits divisible 7
      dbsp ("J7 ");
      dbih (j);
      dbc ('\n');
      bits = l * 7 - j;
      dbsp ("BITS ");
      dbih (j);
      dbc ('\n');
      for (b1 = 0; b1 < bits; b1++)
	set_bit (a1, b1, get_bit (a1, b1 + j));
      a1--;
      *a1 = bits / 7;
    }
    else
    {
      dbsp ("8B\n");
      memcpy (a1, a1 + 1 + i, l - 1 - i);
      a1--;
      *a1 = l - 1 - i;
    }
  }

  memcpy (buf_B (), v, file_len);

  get_inode (ino, file);
  if ((ino->type == INODE_TYPE_FILE
       || ino->type == INODE_TYPE_PROTECTED_KEY_DATA) && ino->text[0] == '*')
  {
    ino->text[0] = '>';
    set_inode (ino, file);
  }

  free (v);
  return buf_B ();
}

u8 archive_ctx (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    dbsp ("archive_ctx_create\n");
    ctx->n_default = 3;
    cd_i (glob_mem->sms);
    glob_mem->tmp_b = glob_mem->sms;
    glob_mem->_pwd = glob_mem->sms;
    glob_mem->is_rules = 0;
    ctx->eD = glob_mem->tmp_b;

    if (glob_mem->activ)
      spider_append_r (ctx, &archive_n_off);
    else
      spider_append_r (ctx, &archive_n_on);
  }
  else if (action == APP_LEAVE)
  {
    dbsp ("archive_ctx_free\n");
    spider_clear_r (ctx);
  }
  return APP_OK;
}

u8 archive_folder (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *res;
    u8 j;
    b32 b, r;
    MCFILE *f;
    Inode *ino = inode ();
    u16 i;

    dbsp ("archive_folder_run\n");
    res =
      get_input (locale (lc_Folder_name), 0, 0x0F, NULL,
		 Q_GET_INPUT_ALPHABET);
    if (res == ENULL)
      return APP_END;		// end
    if (res == NULL)
      return APP_BACK;		// back

    j = res[0] - 1;
    res[2 + j] = '\0';

    cd_i (glob_mem->_pwd);
    if (pwd () == glob_mem->sms)
      b = create_dir (pwd (), res + 2, SORT_SMS_SKIP_1);
    else
      b = create_dir (pwd (), res + 2, SORT_SMS);

    if (b < 0)
      return perror (ERR_NO_SPACE);

    r = create_dir (b, t_r, SORT_SMS);

    if (r < 0)
    {
      delete_inode (b);
      return perror (ERR_NO_SPACE);
    }

    if (pwd () != glob_mem->sms)
    {
      u32 len = get_len (glob_mem->links);
      u16 first_char;

      first_char = 1;

      res[0] = fs_ino_block (pwd ());
      res[1] = fs_ino_block (pwd ()) >> 8;
      res[2] = fs_ino_index (pwd ());
      res[3] = fs_ino_block (b);
      res[4] = fs_ino_block (b) >> 8;
      res[5] = fs_ino_index (b);

      if (len != 0)
      {
	len /= 6;
	f = mc_open (glob_mem->links, FILE_READ);
	for (i = 0; i < len; i++)
	{
	  mc_read (f, i * 6, &first_char, 2);
	  if (first_char = 0)
	    break;
	}
	mc_close (f);
      }

      if (first_char == 0)
      {
	f = mc_open (glob_mem->links, FILE_REWRITE);
	mc_rewrite (f, i * 6, res, 6);
      }
      else
      {
	f = mc_open (glob_mem->links, FILE_APPEND);
	mc_append (f, res, 6);
      }
      mc_close (f);
    }

    cd_i (glob_mem->_pwd);
    return APP_BACK;
  }
  return APP_OK;
}

u8 view_msg (u32 file)
{
  u8 *v, *t, *a1, *res;
  u8 *head = NULL;
  u8 ret;
  u8 i, j;

  dbsp ("archive_msg_view\n");
  db_malloc ();

  v = file_2_buf_B (file, 0, &i);
  if (v == ENULL)
    return APP_END;
  if (v == NULL)
    return APP_BACK;

  t = tpdu_seek (v, T_SMS_AA);
  a1 = get_adn (t, MSISDN_SMS);
#define HEAD_LEN	30
  head = malloc (HEAD_LEN);
  if (head)
  {
    i = 0;
    if (a1)
    {
      while (a1[i] != 0xFF && a1[i] != '\0' && i < HEAD_LEN - 2)
      {
	head[i] = a1[i];
	i++;
      }
      head[i] = '\n';
      i++;
    }

    j = *v;
    if ((j & 0x03) == TP_MTI_DELIVER || (j & 0x03) == TP_MTI_STATUS_REPORT)
    {
      t = tpdu_seek (v, T_SMS_SCTS);
      i = i + sprint_date (t, &head[i], HEAD_LEN - 2 - i);
      head[i] = '\n';
      i++;
    }
    head[i] = '\0';
    i++;
  }
  free (a1);
  t = tpdu_seek (v, T_SMS_DCS);
  res = tpdu_seek (v, T_SMS_UDL);
  dbsp ("XLEN ");
  dbch (*res);
  dbc ('\n');
  dbsp ("XDCS ");
  dbch (*t);
  dbc ('\n');
  res--;
  *res = *t;
  res--;
  *res = STRING_SLV;
  ret = display_text (res, head);
  free (head);
  db_malloc ();
  if (ret == APP_END)
    return APP_END;
  return APP_BACK;
}

u8 archive_msg_view (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    return view_msg (glob_mem->tmp_b);
  }
  return APP_OK;
}

u8 archive_msg_reply (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *res = NULL;
    u8 *t;
    u8 *ms = NULL;
    u8 i;

    dbsp ("archive_msg_reply\n");

    res = file_2_buf_B (glob_mem->tmp_b, 1, &i);
    if (res == ENULL)
      return APP_END;
    if (res == NULL)
      return APP_BACK;
    t = tpdu_seek (res, T_SMS_AA);
    ms = msisdncpy (t, MSISDN_SMS, MEM_R);

    if (i)
      i = MSG_CRYPTED;
    else
      i = MSG_PLAIN;

    i = send_msg (MSG_PLAIN, NULL, ms);
    free (ms);
    return i;
  }
  return APP_OK;
}

u8 i_in_rws (u16 i, u16 * rws, u8 rws_p)
{
  u8 j;

  for (j = 0; j < rws_p; j++)
    if (rws[j] == i)
      return 1;
  return 0;
}

u8 archive_delete (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 res;
    u32 len;

    dbsp ("archive_delete\n");
    Inode *ino = inode ();
    u16 i;
    u8 j;
    u8 *buf = buf_A ();

    get_inode (ino, glob_mem->tmp_b);
    if (ino->type == INODE_TYPE_DIR && !glob_mem->is_rules)
    {
      i = ino->content;
      j = ino->content_index;
      get_inode (ino, fs_ino_p (ino->content, ino->content_index));

      if (ino->next == i && ino->next_index == j)
	get_inode (ino, glob_mem->tmp_b);
      else			// folder not empty
      {
	res = display_text (locale (lc_Not_Empty), NULL);
	if (res == APP_END)
	  return APP_END;
	return APP_BACK;
      }
    }

    res = yesno (ino->text, Q_YESNO_DELETE_QMARK, 1);
    if (res == APP_END)
      return APP_END;
    if (res == APP_YES)
    {
      // dir - list of subdirs

      if (glob_mem->is_rules)
      {
	u8 buf[RULE_POS];
	MCFILE *f;

	dbsp ("archive_delete_rule\n");

	f = mc_open (glob_mem->tmp_b, FILE_READ);
	mc_read (f, RULE_MSISDN, buf, RULE_POS);
	mc_close (f);
	i = (u16) buf[0] | (buf[1] << 8);

	buf[0] = 0;
	f = mc_open (glob_mem->rules, FILE_REWRITE);
	mc_rewrite (f, i * RULE_LEN, buf, 1);
	mc_close (f);
      }

      get_inode (ino, glob_mem->tmp_b);
      if (ino->type == INODE_TYPE_DIR)
      {
	MCFILE *f;
	u16 len;
	u8 pos;
	u8 start;
	u8 row[6];
	u16 *rws = buf_B ();
	u8 rws_p = 0;

	dbsp ("archive_delete_dir\n");

#define MAX_DIR_COUNT	(256/3)
	/*
	   First construct array of dirs to be deleted
	   dir->subdir 3->3 (block, index)->(block, index).
	   Result is in buf.

	   Fill rws with rows to be deleted in /.links,
	   i.e. the rows are no more valid.
	 */

	pos = start = 0;

	buf[pos * 3] = fs_ino_block (glob_mem->tmp_b);
	buf[pos * 3 + 1] = fs_ino_block (glob_mem->tmp_b) >> 8;
	buf[pos * 3 + 2] = fs_ino_index (glob_mem->tmp_b);
	dbsp ("ins rule ");
	dbch (buf[pos * 3]);
	dbch (buf[pos * 3 + 1]);
	dbch (buf[pos * 3 + 2]);
	dbc ('\n');
	pos++;

	len = get_len (glob_mem->links);
	if (len != 0)
	{
	  len /= 6;
	  f = mc_open (glob_mem->links, FILE_REWRITE);
	  do
	  {
	    for (i = 0; i < len; i++)
	    {
	      mc_read (f, i * 6, row, 6);
	      dbsp ("link_L ");
	      dbch (row[0]);
	      dbch (row[1]);
	      dbch (row[2]);
	      dbc ('\n');
	      dbsp ("link_R ");
	      dbch (row[3]);
	      dbch (row[4]);
	      dbch (row[5]);
	      dbc ('\n');

	      if ((row[0] != 0 || row[1] != 0)
		  && i_in_rws (i, rws, rws_p) == 0)
	      {
		if (row[3] == buf[start * 3]
		    && row[4] == buf[start * 3 + 1]
		    && row[5] == buf[start * 3 + 2])
		{
		  rws[rws_p] = i;
		  rws_p++;
		}
		else
		  if (row[0] == buf[start * 3]
		      && row[1] == buf[start * 3 + 1]
		      && row[2] == buf[start * 3 + 2])
		{
		  buf[pos * 3] = row[3];
		  buf[pos * 3 + 1] = row[4];
		  buf[pos * 3 + 2] = row[5];
		  dbsp ("ins rule ");
		  dbch (buf[pos * 3]);
		  dbch (buf[pos * 3 + 1]);
		  dbch (buf[pos * 3 + 2]);
		  dbc ('\n');
		  pos++;
		  if (pos == MAX_DIR_COUNT)
		  {
		    mc_close (f);
		    return perror (ERR_NO_RAM);
		  }
		  rws[rws_p] = i;
		  rws_p++;
		}
	      }
	    }
	    start++;
	  }
	  while (start < pos);

	  /*
	     Invalidate rows.
	   */
	  row[0] = row[1] = 0;
	  for (i = 0; i < rws_p; i++)
	    mc_rewrite (f, rws[i] * 6, row, 6);

	  mc_close (f);
	}

	/*
	   Now go thru rows in /.rules and find rows matching
	   deleted dirs.
	 */
	len = get_len (glob_mem->rules);
	if (len != 0)
	{
	  len /= RULE_LEN;

	  f = mc_open (glob_mem->rules, FILE_REWRITE);
	  for (j = 0; j < pos; j++)
	  {
	    dbsp ("del rule ");
	    dbch (buf[j * 3]);
	    dbch (buf[j * 3 + 1]);
	    dbch (buf[j * 3 + 2]);
	    dbc ('\n');
	    for (i = 0; i < len; i++)
	    {
	      mc_read (f, i * RULE_LEN + RULE_MSISDN, row, 3);
	      dbsp ("row ");
	      dbch (row[0]);
	      dbch (row[1]);
	      dbch (row[2]);
	      dbc ('\n');
	      if (row[0] == buf[3 * j] && row[1] == buf[3 * j + 1]
		  && row[2] == buf[3 * j + 2])
	      {
		dbsp ("del rule X ");
		dbch (buf[3 * j]);
		dbch (buf[3 * j + 1]);
		dbch (buf[3 * j + 2]);
		dbc ('\n');
		row[0] = 0;
		mc_rewrite (f, i * RULE_LEN, row, 1);	// del rule
	      }
	    }
	  }

	  mc_close (f);
	}
      }
      dbsp ("archive_delete_inode\n");

      if (glob_mem->tmp_b == glob_mem->drop)
	glob_mem->drop = 0;
      delete_inode (glob_mem->tmp_b);
      spider_back (ctx);
    }
    return APP_BACK;
  }
  return APP_OK;
}

u8 archive_delete_all (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 res;

    dbsp ("archive_delete_all\n");

    res = yesno (locale (lc_Delete_All), Q_YESNO_QMARK, 1);
    if (res == APP_END)
      return APP_END;
    if (res == APP_YES)
    {
      glob_mem->drop = 0;
      delete_files (glob_mem->_pwd);
    }
    return APP_BACK;
  }
  return APP_OK;
}

u8 archive_move (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    dbsp ("archive_move\n");
    glob_mem->drop = glob_mem->tmp_b;
    return APP_BACK;
  }
  return APP_OK;
}

u8 archive_drop (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    dbsp ("archive_drop\n");
    mc_move (glob_mem->drop, glob_mem->_pwd, SORT_SMS);
    glob_mem->drop = 0;
    return APP_BACK;
  }
  return APP_OK;
}

u8 archive_msg_forward (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    b16 j;
    u8 *res, *v, *t;
    u8 i;

    v = NULL;

    dbsp ("archive_msg_forward\n");
    j = get_len (glob_mem->tmp_b);

    res = file_2_buf_B (glob_mem->tmp_b, 0, &i);
    if (res == ENULL)
      return APP_END;
    if (res == NULL)
      return APP_BACK;

    if (i)
      i = MSG_CRYPTED;
    else
      i = MSG_PLAIN;

    v = malloc (j);
    if (v != NULL)
    {
      memcpy (v, res, j);

      t = tpdu_seek (v, T_SMS_DCS);
      res = tpdu_seek (v, T_SMS_UDL);
      res--;
      *res = *t;
      res--;
      *res = STRING_SLV;
      i = send_msg (i, res, NULL);	// TODO FIX FOR glob_mem->msg_buf
      free (v);
    }
    else
      return perror (ERR_NO_RAM);

    return i;
  }
  return APP_OK;
}

u8 archive_rules (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    dbsp ("archive_rules_ctx\n");
    if (glob_mem->is_rules == 0)
    {
      cd_i (glob_mem->tmp_b);
      cd (t_r);
      glob_mem->_pwd = pwd ();
      glob_mem->is_rules = 1;
      ctx->eD = glob_mem->tmp_b;
    }
    else
    {
      ctx->eD = glob_mem->parent_pwd;
    }
  }
  else if (action == APP_LEAVE)
  {
    dbsp ("archive_rules_free\n");
  }
  return APP_OK;
}

u8 archive_off (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    Pers_mem *p = app_data ();

    wb (&p->activ, 0);
    glob_mem->activ = 0;
    return APP_BACK;
  }

  return APP_OK;
}

u8 archive_on (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    Pers_mem *p = app_data ();

    wb (&p->activ, 1);
    glob_mem->activ = 1;
    return APP_BACK;
  }

  return APP_OK;
}

u8 archive_new_rule (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *res, *buf;
    b32 w, b;
    b32 r, l;
    u16 i;
    u8 first_char;
    u8 alpha[ALPHA_LEN];
    MCFILE *f;
    Inode *ino = inode ();
    u16 parent;
    u8 parent_index;

    dbsp ("archive_new_rule_run\n");
    res = msisdn (alpha);
    if (res == NULL)
      return APP_BACK;
    if (res == ENULL)
      return APP_END;

    cd_i (glob_mem->_pwd);
    get_inode (ino, pwd ());
    parent = ino->parent;
    parent_index = ino->parent_index;

    buf = malloc (RULE_LEN);
    if (buf == NULL)
      return perror (ERR_NO_RAM);

    r = create_file (glob_mem->_pwd, alpha, SORT);
    if (r < 0)
    {
      free (buf);
      return perror (ERR_NO_SPACE);
    }

    bzero (buf, RULE_LEN);
    memcpy (buf, res, res[0] + 1);
    buf[RULE_MSISDN] = parent;
    buf[RULE_MSISDN + 1] = parent >> 8;
    buf[RULE_MSISDN + 2] = parent_index;

    first_char = 0xFF;

    l = get_len (glob_mem->rules);
    l = l / RULE_LEN;

    if (l > 0)
    {
      f = mc_open (glob_mem->rules, FILE_READ);
      for (i = 0; i < l; i++)
      {
	mc_read (f, i * RULE_LEN, &first_char, 1);
	if (first_char == 0)
	  break;
      }
      mc_close (f);
    }

    if (first_char == 0)
    {
      f = mc_open (glob_mem->rules, FILE_REWRITE);
      mc_rewrite (f, i * RULE_LEN, buf, RULE_LEN);
      l = i;
    }
    else
    {
      f = mc_open (glob_mem->rules, FILE_APPEND);
      mc_append (f, buf, RULE_LEN);
    }
    mc_close (f);

    buf[RULE_MSISDN] = l;
    buf[RULE_MSISDN + 1] = l >> 8;

    f = mc_open (r, FILE_APPEND);
    mc_append (f, buf, RULE_LEN);
    mc_close (f);

    free (buf);

    return APP_BACK;
  }
  return APP_OK;
}

u8 file_sms_cb (u32 f, u8 action)
{
  dbsp (">file_app_cb\n");
  if (action == FILE_SELECT)
  {
    return view_msg (f);
  }
  return APP_USE_DEFAULT;
}

SNodeP archive_n = { lc_Archive, archive_ctx };
static SNodeP archive_n_new_folder = { lc_New_Folder, archive_folder };
static SNodeP archive_n_msg_view = { LC_VIEW, archive_msg_view };
static SNodeP archive_n_msg_reply = { lc_Msg_Reply, archive_msg_reply };
static SNodeP archive_n_msg_forward = { lc_Msg_Forward, archive_msg_forward };
static SNodeP archive_n_new_rule = { lc_New_Rule, archive_new_rule };
static SNodeP archive_n_rules = { lc_Rules, archive_rules };
static SNodeP archive_n_delete_folder = { lc_Delete_Folder, archive_delete };
static SNodeP archive_n_delete_msg = { lc_Delete_Msg, archive_delete };
static SNodeP archive_n_delete_rule = { lc_Delete_Rule, archive_delete };
static SNodeP archive_n_delete_all = { lc_Delete_All, archive_delete_all };
static SNodeP archive_n_move = { lc_Move, archive_move };
static SNodeP archive_n_drop = { lc_Drop, archive_drop };
static SNodeP archive_n_off = { lc_Off, archive_off };
static SNodeP archive_n_on = { lc_On, archive_on };

SEdgeP archive_edges_p[] = {
  {&archive_n, &archive_n_new_folder},
  {&archive_n_rules, &archive_n_new_rule},
  NULL
};

#endif
